from cmq.aws.aws import AWSResource
from cmq.aws.resource.keyspace_table import keyspace_table


class keyspace(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "keyspaces"
        self._resource = "keyspace"
        self._list_function = "list_keyspaces"
        self._list_key = "keyspaces"

    def table(self) -> AWSResource:
        return keyspace_table(self)


from importlib import metadata


class ResourceLoader:

    def __getattr__(self, attr: str) -> "ResourcePlugin":
        resource_class = PluginManager().get_resource_class(attr)
        resource_instance = resource_class(self)
        return resource_instance


class ResourcePlugin(ResourceLoader):
    pass


class SessionPlugin(ResourceLoader):
    pass



class PluginManager:

    def __init__(self, group: str = "cmq.provider.aws"):
        self.group = group
        self.sessions: dict[str, SessionPlugin] = {}
        self.resources: dict[str, ResourcePlugin] = {}
        self.load()

    def load(self):
        entry_points = list(metadata.entry_points(group=self.group))
        for ep in entry_points:
            plugin = ep.load()
            if issubclass(plugin, SessionPlugin):
                self.sessions[ep.name] = plugin
            elif issubclass(plugin, ResourcePlugin):
                self.resources[ep.name] = plugin
            else:
                raise ValueError(f"Unknown plugin subclass type: {ep.name}")

    def get_sessions(self) -> dict[str, SessionPlugin]:
        return self.sessions

    def get_resources(self) -> dict[str, ResourcePlugin]:
        return self.resources

    def get_session_class(self, name: str) -> SessionPlugin:
        if name in self.sessions:
            return self.sessions[name]
        raise ValueError(f"Unknown session class {name}")

    def get_resource_class(self, name: str) -> ResourcePlugin:
        if name in self.resources:
            return self.resources[name]
        raise ValueError(f"Unknown resource class {name}")


from cmq.aws.aws import AWSResource


class elasticache_parameter_group(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "elasticache"
        self._resource = "elasticache_parameter_group"
        self._list_function = "describe_cache_parameter_groups"
        self._list_key = "CacheParameterGroups"



from cmq.aws.aws import AWSResource
from cmq.aws.resource.log_stream import log_stream
from cmq.aws.resource.log_event import log_event


class log(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "logs"
        self._resource = "log-group"
        self._list_function = "describe_log_groups"

    @property
    def stream(self) -> log_stream:
        return log_stream(self)

    @property
    def event(self, **kwargs) -> log_event:
        return log_event(self).__call__(**kwargs)


from typing import Any
from cmq.aws.aws import AWSResource
from cmq.aws.resource.elasticache_parameter_group import elasticache_parameter_group


class elasticache(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "elasticache"
        self._resource = "elasticache"
        self._list_function = "describe_cache_clusters"
        self._list_key = "CacheClusters"

        self._tag_function = "list_tags_for_resource"
        self._tag_function_key = "ResourceName"

        self._metric_namespace = "AWS/ElastiCache"
        self._metric_dimension_name = "CacheClusterId"
        self._metric_dimension_resource_key = "CacheClusterId"

    def parameter_group(self) -> AWSResource:
        return elasticache_parameter_group(self)

    def _get_tag_resource_identifier(self, context: dict[str, Any], resource: dict[str, Any]) -> str:
        return resource["ARN"]



import csv
import io
import os
from atpbar import flushing
from benedict import benedict
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Any
from rich.console import Console
from contextlib import nullcontext, suppress
from cmq.plugin import SessionPlugin
from cmq.plugin import ResourcePlugin


class NodeInterface:

    def __init__(self, parent):
        self._child = None
        self._parent = None
        if parent:
            self._parent = parent
            self._parent._child = self

    def __call__(self):
        return self

    def root(self) -> "ResourceInterface":
        return self._parent.root() if self._parent else self

    def _perform_action(self, action):
        context = {"action": action}
        root = self.root()
        root.traverse(context)

    def traverse(self, context):
        raise NotImplementedError

    def _traverse(self, context):
        if self._child:
            self._child.traverse(context)
        else:
            action = context["action"]
            action(context)


class ResourceInterface(NodeInterface):

    def __init__(self, parent):
        super().__init__(parent)
        self._attrs = []
        self._filters = []
        self._transformed_fields = {}
        self._calculated_fields = {}

    def __call__(self):
        return self

    def root(self) -> "ResourceInterface":
        return self._parent.root() if self._parent else self

    def _perform_action(self, action):
        context = {"action": action}
        root = self.root()
        root.traverse(context)

    def traverse(self, context):
        raise NotImplementedError

    def _traverse(self, context):
        if self._child:
            self._child.traverse(context)
        else:
            action = context["action"]
            action(context)

    def attr(self, *args: list[str]) -> "ResourceInterface":
        """
        Adds the given attributes to the resource.

        Args:
            *args (list[str]): The attributes to be added.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._attrs.extend(args)
        return self

    def filter(self, func: callable) -> "ResourceInterface":
        """
        Adds a filter function to the resource.

        Args:
            func (callable): The filter function to be added.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(func)
        return self

    def transform(self, key: str, func: callable) -> "ResourceInterface":
        """
        Adds a transformation function to the resource.

        Args:
            key (str): The key of the attribute to transform.
            func (callable): The transformation function to be added.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._transformed_fields[key] = func
        return self

    def calculate(self, key: str, func: callable) -> "ResourceInterface":
        """
        Adds a calculation function to the resource.

        Args:
            key (str): The new key of the attribute.
            func (callable): The calculation function to be used.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._calculated_fields[key] = func
        return self

    def _transform(self, resource: dict) -> dict:
        if isinstance(resource, dict):
            for key, func in self._transformed_fields.items():
                resource[key] = func(resource.get(key)) if key in resource else None
        return resource

    def _calculate(self, resource: dict) -> dict:
        if isinstance(resource, dict):
            for key, func in self._calculated_fields.items():
                resource[key] = func(resource)
        return resource

    def _get_attr_from_resource(self, resource) -> Any:
        if self._attrs and isinstance(resource, dict):
            return {key: self._safe_key(resource, key) for key in self._attrs}
        else:
            return resource

    def _get_attr(self, resource_list: list) -> list:
        return [self._get_attr_from_resource(resource) for resource in resource_list]

    def _safe_key(self, resource: dict, key: str) -> Any:
        with suppress(AttributeError, KeyError, ValueError):
            return benedict(resource).get(key)
        return None

    def _safe_filter(self, func: callable, resource: dict) -> bool:
        with suppress(AttributeError, KeyError, TypeError):
            return func(resource)
        return False

    def _exclude(self, resources: list) -> list:
        if self._transformed_fields:
            resources = [self._transform(r) for r in resources]
        if self._calculated_fields:
            resources = [self._calculate(r) for r in resources]

        return [r for r in resources if all(self._safe_filter(f, r) for f in self._filters)]

    def eq(self, key, value) -> "ResourceInterface":
        """
        Adds an equality filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (Any): The value to compare against.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key) == value)
        return self

    def ne(self, key, value) -> "ResourceInterface":
        """
        Adds a not-equal filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (Any): The value to compare against.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key) != value)
        return self

    def in_(self, key, value) -> "ResourceInterface":
        """
        Adds an "in" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (list): The list of values to check for inclusion.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key) in value)
        return self

    def contains(self, key, value) -> "ResourceInterface":
        """
        Adds a "contains" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (Any): The value to check for containment.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: value in benedict(x).get(key))
        return self

    def not_contains(self, key, value) -> "ResourceInterface":
        """
        Adds a "not contains" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (Any): The value to check for non-containment.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: value not in benedict(x).get(key))
        return self

    def starts_with(self, key, value) -> "ResourceInterface":
        """
        Adds a "starts with" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (str): The value to check for starting with.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key).startswith(value))
        return self

    def ends_with(self, key, value) -> "ResourceInterface":
        """
        Adds an "ends with" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (str): The value to check for ending with.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key).endswith(value))
        return self

    def gt(self, key, value) -> "ResourceInterface":
        """
        Adds a "greater than" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (Any): The value to compare against.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key) > value)
        return self

    def lt(self, key, value) -> "ResourceInterface":
        """
        Adds a "less than" filter to the resource.

        Args:
            key (str): The key of the attribute to compare.
            value (Any): The value to compare against.

        Returns:
            ResourceInterface: The updated resource object.
        """
        self._filters.append(lambda x: benedict(x).get(key) < value)
        return self

    def context(self, status):
        raise NotImplementedError

    def progress(self, resources):
        raise NotImplementedError


class PagedResourceInterface(ResourceInterface):

    def __init__(self, parent):
        super().__init__(parent)
        self._limit = None

    def _get_pages(self, context) -> Any:
        raise NotImplementedError

    def limit(self, limit: int) -> "PagedResourceInterface":
        """
        Limits the number of resources to be retrieved.

        Args:
            limit (int): The number of resources to be retrieved.

        Returns:
            PagedResourceInterface: The updated resource object.
        """
        self._limit = limit
        return self

    def get_paged_results(self, page) -> list:
        raise NotImplementedError

    def paginate(self, context) -> list:
        resources = []
        for page in self._get_pages(context):
            resources.extend(self.get_paged_results(page))
            if self._limit and len(resources) >= self._limit:
                resources = resources[:self._limit]
                break
        return resources


class Resource(PagedResourceInterface, ResourcePlugin):

    def __init__(self, parent):
        super().__init__(parent)
        self._resource = ""

    def enable_console(self) -> bool:
        return str(os.getenv("CMQ_VERBOSE_OUTPUT", "false")).lower() == "true"

    def list(self) -> list:
        """
        Retrieves a list of resources based on the applied filters.

        Returns:
            list: A list of resource objects.
        """
        results: list = []
        self._perform_action(partial(self._list, results))
        return results

    def dict(self) -> dict:
        """
        Retrieves a dictionary of resources based on the applied filters.

        Returns:
            dict: A dictionary of resource objects.
        """
        results: dict = {}
        self._perform_action(partial(self._dict, results))
        return results

    def csv(self, flat: bool=False) -> str:
        """
        Retrieves a list of resources based on the applied filters and returns them in CSV format.

        Args:
            flat (bool, optional): Specifies whether the dictionaries should be flattened or not. Defaults to False.

        Returns:
            str: List of resources in CSV format.
        """
        results: dict = {}
        self._perform_action(partial(self._dict, results))
        return self._to_csv(results, flat)

    def do(self, action: callable) -> None:
        """
        Performs a custom action on the resources.

        Args:
            action (callable): The action to be performed.
        """
        self._perform_action(action)

    def _list(self, results, context) -> None:
        raise NotImplementedError

    def _dict(self, results, context) -> None:
        raise NotImplementedError

    def _to_csv(self, results, flat: bool) -> str:
        """
        Transform a dictionary of resources into a CSV string in memory
        """
        if not results:
            return ""

        # Get all the keys from the dictionaries
        keys = set({'session'})
        for session, resources in results.items():
            if flat:
                resources = self._flatten(resources)
                results[session] = resources
            keys = self._get_keys_from_dicts(keys, resources)
        keys = sorted(keys)

        # Write the CSV file into memory
        with io.StringIO() as output_file:
            dict_writer = csv.DictWriter(output_file, keys)
            dict_writer.writeheader()
            for session, resources in results.items():
                dict_writer.writerows(map(lambda r: {"session": session, **r}, resources))
            return output_file.getvalue()

    def _get_keys_from_dicts(self, keys, results) -> set:
        for resource in results:
            keys.update(resource.keys())
        return keys

    def _flatten(self, resources):
        def flatten_dict(resource: dict, parent_resource: dict | None = None, parent_key: str = ""):
            parent_resource = parent_resource or {}
            for key in list(resource.keys()):
                value = resource[key]
                new_key = f"{parent_key}.{key}" if parent_key else key
                if isinstance(value, dict):
                    flatten_dict(value, parent_resource, new_key)
                    del resource[key]
                else:
                    parent_resource[new_key] = value
            return parent_resource
        return [flatten_dict(resource) for resource in resources]


class Session(Resource, SessionPlugin):

    console = Console()

    def context(self, status):
        if self.enable_console():
            return Console().status(status)
        else:
            return nullcontext()

    def get_sessions() -> list:
        raise NotImplementedError

    def get_session_context(self, resource: dict) -> dict:
        """
        Get the session context for the resource. This is a dictionary that will be passed to the traverse function.
        It should contain all the information needed to start a client session.
        The dictionary should contain the following:
        - session_resource: The resource dictionary
        - session_name: The name of the session
        - aws_region: The region of the session
        - aws_account: The account of the session
        - aws_session: The boto3 session object
        """
        raise NotImplementedError

    def traverse(self, context):
        functions = []
        for session in self.get_sessions():
            session_context = self.get_session_context(session)
            session_context.update(context)
            functions.append(partial(self._traverse, session_context))

        with flushing(), ThreadPoolExecutor() as executor:
            running_tasks = [executor.submit(task) for task in functions]
            for running_task in running_tasks:
                running_task.result()

    def _list(self, results, context) -> None:
        results.append(context["session_resource"])


from _typeshed import Incomplete

# Manually added resource imports to provide type hints
from cmq.aws.resource.address import address
from cmq.aws.resource.alarm import alarm
from cmq.aws.resource.cloudformation import cloudformation
from cmq.aws.resource.cloudtrail import cloudtrail
from cmq.aws.resource.dynamodb import dynamodb
from cmq.aws.resource.elasticache_parameter_group import elasticache_parameter_group
from cmq.aws.resource.elasticache_subnet_group import elasticache_subnet_group
from cmq.aws.resource.elasticache import elasticache
from cmq.aws.resource.function import function
from cmq.aws.resource.kinesis import kinesis
from cmq.aws.resource.kms import kms
from cmq.aws.resource.log_event import log_event
from cmq.aws.resource.log_stream import log_stream
from cmq.aws.resource.log import log
from cmq.aws.resource.metric import metric
from cmq.aws.resource.rds_parameter_group import rds_parameter_group
from cmq.aws.resource.rds import rds
from cmq.aws.resource.resource_explorer import resource_explorer
from cmq.aws.resource.resource_group import resource_group
from cmq.aws.resource.role import role
from cmq.aws.resource.s3_object import s3_object
from cmq.aws.resource.s3 import s3
from cmq.aws.resource.sns import sns
from cmq.aws.resource.sqs import sqs
from cmq.aws.resource.user import user
from cmq.aws.resource.user_key import user_key


class ResourceInterface:
    def __init__(self, parent) -> None: ...
    def __call__(self): ...
    def root(self) -> ResourceInterface: ...
    def traverse(self, context) -> None: ...
    def config_env(self) -> None: ...
    def context(self, status) -> None: ...
    def progress(self, resources) -> None: ...

    # Useful methods for managing resources
    def attr(self, *args) -> ResourceInterface: ...
    def filter(self, func) -> ResourceInterface: ...
    def transform(self, key, func) -> ResourceInterface: ...
    def eq(self, key, value) -> ResourceInterface: ...
    def ne(self, key, value) -> ResourceInterface: ...
    def in_(self, key, value) -> ResourceInterface: ...
    def contains(self, key, value) -> ResourceInterface: ...
    def not_contains(self, key, value) -> ResourceInterface: ...
    def starts_with(self, key, value) -> ResourceInterface: ...
    def ends_with(self, key, value) -> ResourceInterface: ...
    def gt(self, key, value) -> ResourceInterface: ...
    def lt(self, key, value) -> ResourceInterface: ...

    # Manually added resource to provide type hints
    def address(self) -> address: ...
    def alarm(self) -> alarm: ...
    def cloudformation(self) -> cloudformation: ...
    def cloudtrail(self) -> cloudtrail: ...
    def dynamodb(self) -> dynamodb: ...
    def elasticache_parameter_group(self) -> elasticache_parameter_group: ...
    def elasticache_subnet_group(self) -> elasticache_subnet_group: ...
    def elasticache(self) -> elasticache: ...
    def function(self) -> function: ...
    def kinesis(self) -> kinesis: ...
    def kms(self) -> kms: ...
    def log_event(self) -> log_event: ...
    def log_stream(self) -> log_stream: ...
    def log(self) -> log: ...
    def metric(self) -> metric: ...
    def rds_parameter_group(self) -> rds_parameter_group: ...
    def rds(self) -> rds: ...
    def resource_explorer(self) -> resource_explorer: ...
    def resource_group(self) -> resource_group: ...
    def role(self) -> role: ...
    def s3_object(self) -> s3_object: ...
    def s3(self) -> s3: ...
    def sns(self) -> sns: ...
    def sqs(self) -> sqs: ...
    def user(self) -> user: ...
    def user_key(self) -> user_key: ...
    def ec2(self) -> user_key: ...

class PagedResourceInterface(ResourceInterface):
    def get_paged_results(self, page) -> list: ...
    def paginate(self, context) -> list: ...

class Resource(PagedResourceInterface):
    def list(self) -> list: ...
    def dict(self) -> dict: ...
    def csv(self, flat=False) -> str: ...
    def do(self, action: callable) -> None: ...

class Session(Resource):
    console: Incomplete
    def context(self, status): ...
    def get_sessions() -> list: ...
    def get_session_context(self, resource: dict) -> dict: ...
    def traverse(self, context) -> None: ...



from typing import Any

from cmq.aws.aws import AWSResource
from cmq.aws.resource.s3_object import s3_object


class s3(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "s3"
        self._resource = "bucket"
        self._list_function = "list_buckets"

        self._tag_function = "get_bucket_tagging"
        self._tag_function_key = "Bucket"

        self._metric_namespace = "AWS/S3"
        self._metric_dimension_name = "BucketName"
        self._metric_dimension_resource_key = "Name"

    def objects(self, **kwargs) -> AWSResource:
        return s3_object(self)(**kwargs)

    def _get_tag_resource_identifier(self, context: dict[str, Any], resource: dict[str, Any]) -> str:
        return resource["Name"]

    def _get_pages(self, context) -> Any:
        client = self.get_client(context)
        callable = getattr(client, self._list_function)
        return [callable(**self._list_parameters)]



import json
from datetime import datetime, timedelta, timezone
from cmq.aws.aws import AWSResource


class cloudtrail(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "cloudtrail"
        self._resource = "event"
        self._list_function = "lookup_events"
        self._transformers = {"CloudTrailEvent": json.loads}

    def event_name(self, event_name: str) -> "cloudtrail":
        """
        Sets the event name for filtering events.

        Args:
            event_name (str): The event name to filter by.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        lookup = self._list_parameters.setdefault("LookupAttributes", [])
        lookup.append({"AttributeKey": "EventName", "AttributeValue": event_name})
        return self

    def event_source(self, event_source: str) -> "cloudtrail":
        """
        Sets the event source for filtering events.

        Args:
            event_source (str): The event source to filter by.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        lookup = self._list_parameters.setdefault("LookupAttributes", [])
        lookup.append({"AttributeKey": "EventSource", "AttributeValue": event_source})
        return self

    def resource_name(self, resource_name: str) -> "cloudtrail":
        """
        Sets the resource name for filtering events.

        Args:
            resource_name (str): The resource name to filter by.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        lookup = self._list_parameters.setdefault("LookupAttributes", [])
        lookup.append({"AttributeKey": "ResourceName", "AttributeValue": resource_name})
        return self

    def user_name(self, user_name: str) -> "cloudtrail":
        """
        Sets the user name for filtering events.

        Args:
            user_name (str): The user name to filter by.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        lookup = self._list_parameters.setdefault("LookupAttributes", [])
        lookup.append({"AttributeKey": "Username", "AttributeValue": user_name})
        return self

    def access_key(self, access_key: str) -> "cloudtrail":
        """
        Sets the access key for filtering events.

        Args:
            access_key (str): The access key to filter by.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        lookup = self._list_parameters.setdefault("LookupAttributes", [])
        lookup.append({"AttributeKey": "AccessKeyId", "AttributeValue": access_key})
        return self

    def event_time(self, start_time: datetime, end_time: datetime) -> "cloudtrail":
        """
        Sets the event time range for filtering events.

        Args:
            start_time (datetime): The start time of the event range.
            end_time (datetime): The end time of the event range.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        self._list_parameters.update({"StartTime": start_time, "EndTime": end_time})
        return self

    def last(self, **kwargs: dict[str, int]) -> "cloudtrail":
        """
        Sets the number of events to retrieve.

        Args:
            **kwargs: The keyword arguments for the timedelta.

        Returns:
            cloudtrail: The cloudtrail object for method chaining.
        """
        self._list_parameters.update({"StartTime": datetime.now(tz=timezone.utc) - timedelta(**kwargs),
                                      "EndTime": datetime.now(tz=timezone.utc)})
        return self


import boto3
import botocore
import os
import sys
import configparser

from typing import Any
from cmq.base import Session


class profile(Session):

    def __init__(self, parent: object=None, name: str | None = None, **kwargs: dict[str, Any]):
        """
        Initialize a new Profile object.

        Args:
            parent (object): The parent object. Defaults to None.
            name (str | None): The name of the profile. Defaults to None.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(parent)
        self._name = name
        self._list_attr = ["id", "name", "region"]
        self._list_parameters = kwargs
        self.config_path = os.path.expanduser(os.getenv("AWS_CONFIG_FILE", "~/.aws/config"))

    def get_profiles(self) -> tuple:
        config = configparser.ConfigParser()
        config.read([self.config_path])
        return config, config.sections() if self._name is None else [f"profile {self._name}"]

    def get_profile_name(self, profile: str) -> str:
        return profile.split()[-1] if profile.startswith("profile ") else profile

    def get_session_context(self, profile: dict) -> dict:
        return {
            "session_resource": profile,
            "session_name": profile["name"],
            "aws_region": profile.get("region"),
            "aws_session": self.get_session(profile),
        }

    def get_sessions(self):
        with self.context("[bold green]Getting profile list..."):
            config, profiles = self.get_profiles()
            sessions = [{"name": self.get_profile_name(profile), **config[profile]} for profile in profiles]
            return self._get_attr(self._exclude(sessions))

    def get_session(self, profile: dict) -> boto3.Session:
        try:
            return boto3.Session(profile_name=profile["name"])
        except botocore.exceptions.ProfileNotFound:
            profile_name = profile["name"]
            sys.stderr.write(f"Profile {profile_name} not found.\n")

    def _list(self, results, context) -> None:
        results.append(context["session_resource"])

    def _dict(self, results, context) -> None:
        profile_list = results.setdefault(context["session_name"], [])
        profile_list.append(context["session_resource"])



from cmq.aws.aws import AWSResource


class cloudformation(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "cloudformation"
        self._resource = "stack"
        self._list_function = "list_stacks"



from typing import Any
from cmq.aws.aws import AWSResource
from cmq.aws.resource.kms_alias import kms_alias


class kms(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "kms"
        self._resource = "kms"
        self._list_function = "list_keys"

        self._tag_function = "list_resource_tags"
        self._tag_function_key = "KeyId"

        self._describe_function = "describe_key"
        self._describe_function_key = "KeyId"
        self._describe_resource_key = "KeyId"

    def _get_tag_resource_identifier(self, context: dict[str, Any], resource) -> str:
        return resource["KeyId"]

    def _format_tags(self, tags) -> dict:
        return {"Tags": {tag["TagKey"]: tag["TagValue"] for tag in tags}}

    def aliases(self) -> AWSResource:
        return kms_alias(self)


from cmq.aws.aws import AWSResource


class keyspace_table(AWSResource):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._service = "keyspaces"
        self._resource = "table"
        self._list_function = "list_tables"
        self._list_key = "tables"

        self._describe_function = "get_table"

    def get_parameters(self, context: dict) -> dict:
        parameters = self._list_parameters.copy()
        if 'keyspaces_keyspace' in context:
            parameters["keyspaceName"] = context["keyspaces_keyspace"]["keyspaceName"]
        return parameters

    def _describe(self, context, client, resource):
        describe_function = getattr(client, self._describe_function)
        details = describe_function(keyspaceName=resource["keyspaceName"],
                                    tableName=resource["tableName"])
        resource.update({"Describe": details})